#ifndef MATERIAL_H_
#define MATERIAL_H_

#include "vector3.h"
#include "texture.h"

/*! \def NO_TEXTURES
\brief Maximln poet textur piazench materilu.
*/
#define NO_TEXTURES 4

/*! \def IOR_AIR
\brief Index lomu vzduchu za normlnho tlaku.
*/
#define IOR_AIR 1.000293f

/*! \def IOR_AIR
\brief Index lomu vody.
*/
#define IOR_WATER 1.33f

/*! \def IOR_GLASS
\brief Index lomu skla.
*/
#define IOR_GLASS 1.5f

/*! \class Material
\brief A simple material.

\author Tom Fabin
\version 0.9
\date 2011-2018
*/
class Material
{
public:	
	//! Implicitn konstruktor.
	/*!
	Inicializuje vechny sloky materilu na vchoz matn ed materil.
	*/
	Material();

	//! Specializovan konstruktor.
	/*!
	Inicializuje materil podle zadanch hodnot parametr.

	\param name nzev materilu.
	\param ambient barva prosted.
	\param diffuse barva rozptylu.
	\param specular barva odrazu.
	\param emission  barva emise.
	\param shininess lesklost.
	\param ior index lomu.
	\param textures pole ukazatel na textury.
	\param no_textures dlka pole \a textures. Maximln \a NO_TEXTURES - 1.
	*/
	Material( std::string & name, const Vector3 & ambient, const Vector3 & diffuse,
		const Vector3 & specular, const Vector3 & emission, const float reflectivity,
		const float shininess, const float ior,
		Texture ** textures = NULL, const int no_textures = 0 );

	//! Destruktor.
	/*!
	Uvoln vechny alokovan zdroje.
	*/
	~Material();

	//void Print();

	//! Nastav nzev materilu.
	/*!	
	\param name nzev materilu.
	*/
	void set_name( const char * name );

	//! Vrt nzev materilu.
	/*!	
	\return Nzev materilu.
	*/
	std::string get_name() const;

	//! Nastav texturu.
	/*!	
	\param slot slo slotu, do kterho bude textura piazena. Maximln \a NO_TEXTURES - 1.
	\param texture ukazatel na texturu.
	*/
	void set_texture( const int slot, Texture * texture );

	//! Vrt texturu.
	/*!	
	\param slot slo slotu textury. Maximln \a NO_TEXTURES - 1.
	\return Ukazatel na zvolenou texturu.
	*/
	Texture * get_texture( const int slot ) const;

public:
	Vector3 ambient; /*!< RGB barva prosted \f$\left<0, 1\right>^3\f$. */
	Vector3 diffuse; /*!< RGB barva rozptylu \f$\left<0, 1\right>^3\f$. */
	Vector3 specular; /*!< RGB barva odrazu \f$\left<0, 1\right>^3\f$. */

	Vector3 emission; /*!< RGB barva emise \f$\left<0, 1\right>^3\f$. */

	float shininess; /*!< Koeficient lesklosti (\f$\ge 0\f$). m je hodnota vt, tm se jev povrch lesklej. */

	float reflectivity; /*!< Koeficient odrazivosti. */
	float ior; /*!< Index lomu. */

	static const char kDiffuseMapSlot; /*!< slo slotu difuzn textury. */
	static const char kSpecularMapSlot; /*!< slo slotu spekulrn textury. */
	static const char kNormalMapSlot; /*!< slo slotu normlov textury. */
	static const char kOpacityMapSlot; /*!< slo slotu transparentn textury. */

private:
	Texture * textures_[NO_TEXTURES]; /*!< Pole ukazatel na textury. */
	/*
	slot 0 - diffuse map + alpha
	slot 1 - specular map + opaque alpha
	slot 2 - normal map
	slot 3 - transparency map
	*/
	
	std::string name_; /*!< Nzev materilu. */
};

#endif
